from typing import Any, Sequence

import attr
from multimethod import multimethod

from visions.relations import IdentityRelation, InferenceRelation, TypeRelation
from visions.types.object import Object
from visions.types.string import String
from visions.types.type import VisionsBaseType


@attr.s(slots=True)
class FQDA:
    local = attr.ib()
    fqdn = attr.ib()

    @staticmethod
    def from_str(s):
        return _to_email(s)


def _to_email(s) -> FQDA:
    if isinstance(s, FQDA):
        return s
    elif isinstance(s, str):
        return FQDA(*s.split("@", maxsplit=1))
    else:
        raise TypeError("Only strings supported")


class EmailAddress(VisionsBaseType):
    """**EmailAddress** implementation of :class:`visions.types.type.VisionsBaseType`.

    Notes:
        The email address should be a **fully qualified domain address** (FQDA)
        FQDA = local part + @ + fully qualified domain name (FQDN)
        This type

    Examples:
        >>> import visions
        >>> x = [FQDA('example','gmail.com'), FQDA.from_str('example@protonmail.com')]
        >>> x in visions.EmailAddress
        True
    """

    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        relations = [
            IdentityRelation(Object),
            InferenceRelation(String),
        ]
        return relations

    @staticmethod
    @multimethod
    def contains_op(item: Any, state: dict) -> bool:
        pass



from typing import Any, Callable, Optional, TypeVar

import attr
from multimethod import multimethod

T = TypeVar("T")


def func_repr(func: Callable) -> str:
    return func.__name__ if hasattr(func, "__name__") else "lambda"


def identity_transform(series: Any, state: dict = dict()) -> Any:
    return series


def default_relation(series: Any, state: dict = dict()) -> bool:
    raise NotImplementedError


@attr.s(frozen=True)
class TypeRelation:
    """Relationship encoder between implementations of :class:`visions.types.type.VisionsBaseType`

    Defines a one to one relationship between two :class:`visions.types.type.VisionsBaseType` implementations,
    A and B, with respect to an underlying data series. In order to define a relationship we need
    two methods:

        - **is_relationship**, determines whether a series of type B can be alternatively represented as type A.
        - **transform**, provides a mechanism to convert the series from B -> A.

    For example, the series `pd.Series([1.0, 2.0, 3.0])` is encoded as a sequence of
    floats but in reality they are all integers.

    Examples:
        >>> from visions.types import Integer, Float
        >>> x = pd.Series([1.0, 2.0, 3.0])
        >>> state = dict()
        >>> relation = TypeRelation(Integer, Float)
        >>> relation.is_relation(x, state)
        True

        >>> relation.transform(x, state)
        pd.Series([1, 2, 3])
    """

    related_type = attr.ib()
    inferential: bool = attr.ib()
    transformer: Callable[[T, dict], T] = attr.ib(
        converter=multimethod, repr=func_repr  # type: ignore
    )
    relationship: Callable[[Any, dict], bool] = attr.ib(
        default=default_relation, converter=multimethod, repr=func_repr  # type: ignore
    )
    type = attr.ib(default=None)

    def is_relation(self, series: Any, state: Optional[dict] = None) -> bool:
        if state is None:
            state = {}
        return self.relationship(series, state)

    def transform(self, series: T, state: Optional[dict] = None) -> T:
        if state is None:
            state = {}
        return self.transformer(series, state)

    def __str__(self):
        return f"{self.related_type}->{self.type}"


@attr.s(frozen=True)
class IdentityRelation(TypeRelation):
    relationship: Callable[[T, dict], bool] = attr.ib(repr=func_repr, default=None)
    transformer: Callable[[T, dict], T] = attr.ib(
        default=identity_transform, repr=func_repr
    )
    inferential: bool = attr.ib(default=False)


@attr.s(frozen=True)
class InferenceRelation(TypeRelation):
    relationship: Callable[[T, dict], bool] = attr.ib(
        repr=func_repr, default=default_relation
    )
    transformer: Callable[[T, dict], T] = attr.ib(
        repr=func_repr, default=identity_transform
    )
    inferential: bool = attr.ib(default=True)



from typing import Sequence

from visions.backends.python.series_utils import sequence_not_empty
from visions.backends.python.types.float import no_leading_zeros
from visions.types.complex import Complex
from visions.types.string import String


@Complex.register_relationship(String, Sequence)
def string_is_complex(sequence: Sequence, state: dict) -> bool:
    try:
        coerced = list(string_to_complex(sequence, state))
        return no_leading_zeros(sequence, [r.real for r in coerced])
    except (ValueError, TypeError, AttributeError):
        return False


@Complex.register_transformer(String, Sequence)
def string_to_complex(sequence: Sequence, state: dict) -> Sequence:
    return list(map(complex, sequence))


@Complex.contains_op.register
@sequence_not_empty
def complex_contains(sequence: Sequence, state: dict) -> bool:
    return all(isinstance(value, complex) for value in sequence)



from typing import Any, Sequence

from multimethod import multimethod

from visions.relations import IdentityRelation, TypeRelation
from visions.types.generic import Generic
from visions.types.type import VisionsBaseType


class Sparse(VisionsBaseType):
    """**Sparse** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> import pandas as pd
        >>> import visions
        >>> x = pd.Sparse(pd.Series([np.complex(0, 0), np.complex(1, 2), np.complex(3, -1)]))
        >>> x in visions.Sparse
        True
    """

    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        relations = [IdentityRelation(Generic)]
        return relations

    @staticmethod
    @multimethod
    def contains_op(item: Any, state: dict) -> bool:
        pass



from typing import Any, Dict, Iterable, Optional, Sequence, Set, Tuple, Type

import networkx as nx
import pandas as pd
import pytest

from visions import VisionsBaseType, VisionsTypeset

T = Type[VisionsBaseType]


def is_iter(v: Any) -> bool:
    return isinstance(v, Iterable) and not isinstance(v, (str, bytes))


def sequences_equal(s1: Sequence, s2: Sequence) -> bool:
    for v1, v2 in zip(s1, s2):
        if is_iter(v1) and is_iter(v2):
            if not sequences_equal(v1, v2):
                return False
        elif not (pd.isna(v1) and pd.isna(v2)) and not v1 == v2:
            return False

    return True


def all_series_included(
    series_list: Dict[str, Sequence], series_map: Dict[T, Set[str]]
):
    """Check that all names are indeed used"""
    used_names = {name for names in series_map.values() for name in names}
    names = set(series_list.keys())

    if not names == used_names:
        unused = names - used_names
        not_provided = used_names - names
        # TODO: warning?
        if len(unused) > 0:
            raise ValueError(f"{len(unused)} series not included in tests {unused}")
        if len(not_provided) > 0:
            raise ValueError(
                f"{len(not_provided)} series are included, not not provided {not_provided}"
            )


def get_contains_cases(
    _test_suite: Dict[str, Sequence],
    _series_map: Dict[T, Set[str]],
    typeset: VisionsTypeset,
):
    """Parametrize contains tests

    Args:
        _test_suite: mapping from sequence identifiers to sequences
        _series_map: mapping from type to a set of sequence identifiers
        typeset: A VisionsTypeset

    Returns:
        the args for the generated tests
    """

    # Include children's series in parent
    reversed_topological_edge_sort = list(
        reversed(list(nx.topological_sort(nx.line_graph(typeset.base_graph))))
    )
    for parent, child in reversed_topological_edge_sort:
        _series_map[parent] |= _series_map[child]

    all_series_included(_test_suite, _series_map)

    argsvalues = []
    for name, item in _test_suite.items():
        for type, series_list in _series_map.items():
            args: Dict[str, Any] = {"id": f"{name} x {type}"}

            member = name in series_list
            argsvalues.append(pytest.param(name, item, type, member, **args))

    return {
        "argnames": ["name", "series", "contains_type", "member"],
        "argvalues": argsvalues,
    }


def contains(name: str, series: Sequence, type: T, member: bool) -> Tuple[bool, str]:
    return (
        member == (series in type),
        f"{name} in {type}; expected {member}, got {series in type}",
    )


def get_inference_cases(
    _test_suite: Dict[str, Sequence],
    inferred_series_type_map: Dict[str, T],
    typeset: VisionsTypeset,
) -> Dict[str, Any]:
    argsvalues = []
    for name, series in _test_suite.items():
        if name not in inferred_series_type_map:
            raise ValueError(
                f"{name} has no defined inference type, please add one to the test case mapping"
            )

        expected_type = inferred_series_type_map[name]
        for test_type in typeset.types:
            expected = test_type == expected_type
            args: Dict[str, Any] = {"id": f"{name} x {test_type} expected {expected}"}
            difference = test_type != expected_type
            argsvalues.append(
                pytest.param(name, series, test_type, typeset, difference, **args)
            )
    return {
        "argnames": "name,series,inference_type,typeset,difference",
        "argvalues": argsvalues,
    }


def infers(
    name: str,
    series: Sequence,
    expected_type: T,
    typeset: VisionsTypeset,
    difference: bool,
) -> Tuple[bool, str]:
    from visions.typesets.typeset import get_type_from_path

    _, paths, _ = typeset.infer(series)
    inferred_type = get_type_from_path(paths)

    # inferred_type = typeset.infer_type(series)
    return (
        (inferred_type == expected_type) != difference,
        f"inference of {name} expected {expected_type} to be {not difference} (typeset={typeset}). Path: {paths}",
    )
    # return series in inferred_type, f"series should be member of inferred type"


def all_relations_tested(series_map, typeset):
    # Convert data structure for mapping
    series_map_lookup = {}
    for map_to_type, map_from_type, items in series_map:
        try:
            series_map_lookup[map_to_type][map_from_type] = items
        except KeyError:
            series_map_lookup[map_to_type] = {map_from_type: items}

    missing_relations = set()
    for node in typeset.types:
        for relation in node.relations:
            from_type, to_type = relation.related_type, relation.type
            if relation.inferential and (
                to_type not in series_map_lookup
                or from_type not in series_map_lookup[to_type]
                or len(series_map_lookup[to_type][from_type]) == 0
            ):
                missing_relations.add(str(relation))

    if len(missing_relations) > 0:
        raise ValueError(
            f"Not all inferential relations are tested {missing_relations}"
        )


def get_convert_cases(_test_suite, _series_map, typeset):
    all_relations_tested(_series_map, typeset)

    argsvalues = []
    for name, item in _test_suite.items():
        for source_type, relation_type, series_list in _series_map:
            for namex in series_list:
                if namex not in _test_suite.keys():
                    raise ValueError(
                        f"{namex} specified in convert_map, but not in provided sequences"
                    )

            if item in relation_type:
                args: Dict[str, Any] = {
                    "id": f"{name}: {relation_type} -> {source_type}"
                }
                member = name in series_list
                argsvalues.append(
                    pytest.param(name, source_type, relation_type, item, member, **args)
                )

    return dict(
        argnames=["name", "source_type", "relation_type", "series", "member"],
        argvalues=argsvalues,
    )


def convert(
    name: str, source_type: T, relation_type: T, series: Sequence, member: bool
) -> Tuple[bool, str]:
    relation = source_type.relations.get(relation_type, None)
    is_relation = False if relation is None else relation.is_relation(series, {})

    if not member:
        return (
            (not is_relation),
            f"{source_type}, {relation}, {member}, {name}, {series}",
        )
    else:
        # Note that the transformed series is not exactly the cast series
        transformed_series = list(relation.transform(series, {}))

        return (
            is_relation,
            f"Relationship {relation} for {series} tested false (but shouldn't have). "
            f"Transform result would have been {transformed_series}",
        )


def get_cast_cases(_test_suite: Dict[str, Sequence], _results: Dict) -> Dict:
    argsvalues = []
    for name, item in _test_suite.items():
        changed = name in _results
        value = _results.get(name, "")
        args: Dict[str, Any] = {"id": f"{name}: {changed}"}
        argsvalues.append(pytest.param(name, item, value, **args))

    return dict(
        argnames=["name", "series", "expected"],
        argvalues=argsvalues,
    )


def cast(
    name: str,
    series: Sequence,
    typeset: VisionsTypeset,
    expected: Optional[pd.Series] = None,
) -> Tuple[bool, str]:
    result = typeset.cast_to_inferred(series)
    # TODO: if error also print Path
    if expected is None:
        v = sequences_equal(result, series)
        m = f"Series {name} cast expected {series} (no casting) got {result}"

        if v:
            v = id(series) == id(result)
            m = f"Series {name} memory addresses are not equal, while return value was"
    else:
        v = sequences_equal(result, expected)
        m = f"Series {name} cast expected {expected} got {result}"

    return v, m



# import pandas as pd
#
# from visions import String
# from visions.backends.pandas_be import test_utils
# from visions.relations import InferenceRelation
#
#
# def to_datetime_year_week(series: pd.Series) -> pd.Series:
#     """Convert a series of the format YYYY/UU (year, week) to datetime.
#     A '0' is added as day dummy value, as pandas requires a day value to parse.
#
#     Args:
#         series: the Series to parse
#
#     Returns:
#         A datetime series
#
#     Examples:
#         >>> series = pd.Series(['2018/47', '2018/12', '2018/03'])
#         >>> parsed_series = to_datetime_year_week(series)
#         >>> print(parsed_series.dt.week)
#         0    47
#         1    12
#         2     3
#         dtype: int64
#     """
#     return pd.to_datetime(series + "0", format="%Y/%U%w")
#
#
# def to_datetime_year_month_day(series: pd.Series) -> pd.Series:
#     """Convert a series of the format YYYYMMDD (year, month, day) to datetime.
#
#     Args:
#         series: the Series to parse
#
#     Returns:
#         A datetime series
#
#     Examples:
#         >>> series = pd.Series(['20181201', '20181202', '20181203'])
#         >>> parsed_series = to_datetime_year_week(series)
#         >>> print(parsed_series.dt.day)
#         0    1
#         1    2
#         2    3
#         dtype: int64
#     """
#     return pd.to_datetime(series, format="%Y%m%d")
#
#
# def _to_datetime(func) -> InferenceRelation:
#     return InferenceRelation(
#         relationship=test_utils.coercion_test(func),
#         transformer=func,
#         related_type=String,
#     )
#
#
# def string_to_datetime_year_week() -> InferenceRelation:
#     return _to_datetime(to_datetime_year_week)
#
#
# def string_to_datetime_year_month_day() -> InferenceRelation:
#     return _to_datetime(to_datetime_year_month_day)



from typing import Any, Sequence

from multimethod import multimethod

from visions.relations import IdentityRelation, InferenceRelation, TypeRelation
from visions.types.generic import Generic
from visions.types.string import String
from visions.types.type import VisionsBaseType


class Complex(VisionsBaseType):
    """**Complex** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> x = [complex(0, 0), complex(1, 2), complex(3, -1)]
        >>> x in visions.Complex
        True
    """

    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        relations = [
            IdentityRelation(Generic),
            InferenceRelation(String),
        ]
        return relations

    @staticmethod
    @multimethod
    def contains_op(item: Any, state: dict) -> bool:
        pass



import pandas as pd
from pandas.api import types as pdt

from visions.backends.pandas.series_utils import series_not_empty, series_not_sparse
from visions.types.numeric import Numeric


@Numeric.contains_op.register
@series_not_sparse
@series_not_empty
def numeric_contains_op(series: pd.Series, state: dict) -> bool:
    return pdt.is_numeric_dtype(series)



from datetime import date, time
from typing import Sequence

from visions.types.date import Date
from visions.types.date_time import DateTime


@Date.register_relationship(DateTime, Sequence)
def datetime_is_date(sequence: Sequence, state: dict) -> bool:
    value = time(0, 0)
    return all(v == value for v in sequence)


@Date.register_transformer(DateTime, Sequence)
def datetime_to_date(sequence: Sequence, state: dict) -> Sequence:
    return tuple(map(lambda v: v.date(), sequence))


@Date.contains_op.register
def date_contains(sequence: Sequence, state: dict) -> bool:
    return all(isinstance(value, date) for value in sequence)



from pyspark.sql.dataframe import DataFrame
from pyspark.sql.types import ArrayType, DateType, MapType, StringType, StructType

from visions.types.object import Object


@Object.contains_op.register
def object_contains(sequence: DataFrame, state: dict) -> bool:
    if len(sequence.schema) != 1:
        return False

    dtype = sequence.schema[0].dataType
    return isinstance(dtype, (StringType, DateType, ArrayType, MapType, StructType))



from typing import Any, Sequence

from multimethod import multimethod

from visions.relations import IdentityRelation, InferenceRelation, TypeRelation
from visions.types.object import Object
from visions.types.string import String
from visions.types.type import VisionsBaseType


class Geometry(VisionsBaseType):
    """**Geometry** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> from shapely import wkt
        >>> x = pd.Series([wkt.loads('POINT (-92 42)'), wkt.loads('POINT (-92 42.1)'), wkt.loads('POINT (-92 42.2)')]
        >>> x in visions.geometry
        True
    """

    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        relations = [
            IdentityRelation(Object),
            InferenceRelation(String),
        ]
        return relations

    @staticmethod
    @multimethod
    def contains_op(item: Any, state: dict) -> bool:
        pass



from typing import Dict, List, Tuple, Type

import networkx as nx
import pandas as pd

from visions.types.type import VisionsBaseType
from visions.typesets.typeset import traverse_graph, traverse_graph_with_series

T = Type[VisionsBaseType]


@traverse_graph.register(pd.Series)
def _traverse_graph_series(
    series: pd.Series, root_node: T, graph: nx.DiGraph
) -> Tuple[pd.Series, List[T], dict]:
    return traverse_graph_with_series(root_node, series, graph)


@traverse_graph.register(pd.DataFrame)
def _traverse_graph_dataframe(
    df: pd.DataFrame, root_node: T, graph: nx.DiGraph
) -> Tuple[pd.DataFrame, Dict[str, List[T]], Dict[str, dict]]:
    inferred_values = {
        col: traverse_graph(df[col], root_node, graph) for col in df.columns
    }

    inferred_series = {}
    inferred_paths: Dict[str, List[T]] = {}
    inferred_states: Dict[str, dict] = {}
    for col, (inf_series, inf_path, inf_state) in inferred_values.items():
        assert isinstance(inf_path, list)  # Placate the MyPy Gods

        inferred_series[col] = inf_series
        inferred_paths[col] = inf_path
        inferred_states[col] = inf_state

    return pd.DataFrame(inferred_series), inferred_paths, inferred_states



from ipaddress import _BaseAddress, ip_address

import pandas as pd

from visions.backends.pandas import test_utils
from visions.backends.pandas.series_utils import series_handle_nulls, series_not_empty
from visions.backends.shared.parallelization_engines import pandas_apply
from visions.types.ip_address import IPAddress
from visions.types.string import String


@IPAddress.register_relationship(String, pd.Series)
def string_is_ip_address(series: pd.Series, state: dict) -> bool:
    return test_utils.coercion_test(lambda s: pandas_apply(s, ip_address))(series)


@IPAddress.register_transformer(String, pd.Series)
def string_to_ip_address(series: pd.Series, state: dict) -> pd.Series:
    return pandas_apply(series, ip_address)


@IPAddress.contains_op.register
@series_not_empty
@series_handle_nulls
def ip_address_contains(series: pd.Series, state: dict) -> bool:
    return all(isinstance(x, _BaseAddress) for x in series)



import numpy as np

from visions.backends.numpy.array_utils import array_handle_nulls
from visions.types.float import Float
from visions.types.integer import Integer


@Integer.register_relationship(Float, np.ndarray)
def float_is_integer(series: np.ndarray, state: dict) -> bool:
    return np.all(np.mod(series[~np.isnan(series)], 1) == 0)


# TODO: The array_handle_nulls is actually removing nulls from the result. This is _far_ from ideal but there is no
# other native way to represent nullable integers in numpy
@Integer.register_transformer(Float, np.ndarray)
@array_handle_nulls
def float_to_integer(series: np.ndarray, state: dict) -> np.ndarray:
    return series.astype(int)


@Integer.contains_op.register
@array_handle_nulls
def integer_contains(sequence: np.ndarray, state: dict) -> bool:
    if sequence.shape[0] == 0 or np.issubdtype(sequence.dtype, np.timedelta64):
        return False
    elif np.issubdtype(sequence.dtype, np.integer):
        return True
    elif np.issubdtype(sequence.dtype, np.object_):
        return all(isinstance(v, int) and not isinstance(v, bool) for v in sequence)

    return False



from typing import Any, Sequence

from multimethod import multimethod

from visions.relations import IdentityRelation, InferenceRelation, TypeRelation
from visions.types.object import Object
from visions.types.string import String
from visions.types.type import VisionsBaseType


class URL(VisionsBaseType):
    """**Url** implementation of :class:`visions.types.type.VisionsBaseType`.

    Examples:
        >>> from urllib.parse import urlparse
        >>> urls = ['http://www.cwi.nl:80/%7Eguido/Python.html', 'https://github.com/pandas-profiling/pandas-profiling']
        >>> x = [urlparse(url) for url in urls]
        >>> x in visions.URL
        True
    """

    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        relations = [
            IdentityRelation(Object),
            InferenceRelation(String),
        ]
        return relations

    @staticmethod
    @multimethod
    def contains_op(item: Any, state: dict) -> bool:
        pass



"""Core functionality"""

from visions import types, typesets, utils
from visions.backends import *
from visions.declarative import create_type
from visions.functional import (
    cast_to_detected,
    cast_to_inferred,
    detect_type,
    infer_type,
)
from visions.types import *
from visions.typesets import *

__version__ = "0.7.6"



import os
import sys

import pandas as pd

from visions.backends.pandas.series_utils import series_handle_nulls, series_not_empty
from visions.types.geometry import Geometry
from visions.types.string import String


# TODO: Evaluate https://jorisvandenbossche.github.io/blog/2019/08/13/geopandas-extension-array-refactor/
@Geometry.register_relationship(String, pd.Series)
def string_is_geometry(sequence: pd.Series, state: dict) -> bool:
    """Shapely logs failures at a silly severity, just trying to suppress it's output on failures."""
    from shapely import wkt
    from shapely.errors import WKTReadingError

    # only way to get rid of sys output when wkt.loads hits a bad value
    # TODO: use coercion wrapper for this
    sys.stderr = open(os.devnull, "w")
    try:
        result = all(wkt.loads(value) for value in sequence)
    except (WKTReadingError, AttributeError, UnicodeEncodeError, TypeError):
        result = False
    finally:
        sys.stderr = sys.__stderr__
    return result


@Geometry.register_transformer(String, pd.Series)
def string_to_geometry(series: pd.Series, state: dict) -> pd.Series:
    from shapely import wkt

    return pd.Series([wkt.loads(value) for value in series])


@Geometry.contains_op.register
@series_not_empty
@series_handle_nulls
def geometry_contains(series: pd.Series, state: dict) -> bool:
    from shapely.geometry.base import BaseGeometry

    return all(issubclass(type(x), BaseGeometry) for x in series)



import warnings
from functools import singledispatch
from pathlib import Path
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import networkx as nx
import pandas as pd

from visions.types.generic import Generic
from visions.types.type import VisionsBaseType

TypeOrTypeset = TypeVar("TypeOrTypeset", Type[VisionsBaseType], "VisionsTypeset")
pathTypes = TypeVar(
    "pathTypes", Type[VisionsBaseType], Dict[str, Type[VisionsBaseType]]
)
pdT = TypeVar("pdT", pd.Series, pd.DataFrame)
T = Type[VisionsBaseType]


def build_graph(nodes: Set[Type[VisionsBaseType]]) -> Tuple[nx.DiGraph, nx.DiGraph]:
    """Constructs a traversable relation graph between visions types

    Builds a type relation graph from a collection of :class:`visions.types.type.VisionsBaseType` where
    each node corresponds to a type and each edge is a relation defined on the type.

    Args:
        nodes:  An Sequence of :class:`visions.types.type.VisionsBaseType`

    Returns:
        A directed graph of type relations for the provided nodes.
    """

    style_map = {True: "dashed", False: "solid"}
    relation_graph = nx.DiGraph()
    relation_graph.add_nodes_from(nodes)

    noninferential_edges = []

    for node in nodes:
        for relation in node.relations:
            if relation.related_type not in nodes:
                warnings.warn(
                    f"Provided relations included mapping from {relation.related_type} to {relation.type} "
                    f"but {relation.related_type} was not included in the provided list of nodes"
                )
            else:
                relation_graph.add_edge(
                    relation.related_type,
                    relation.type,
                    relationship=relation,
                    style=style_map[relation.inferential],
                )

                if not relation.inferential:
                    noninferential_edges.append((relation.related_type, relation.type))

    check_graph_constraints(relation_graph)

    base_graph = relation_graph.edge_subgraph(noninferential_edges)
    return relation_graph, base_graph


def check_graph_constraints(relation_graph: nx.DiGraph) -> None:
    """Validates a relation_graph is appropriately constructed

    Args:
        relation_graph: A directed graph representing the set of relations between type nodes.

    """
    check_isolates(relation_graph)
    check_cycles(relation_graph)


def check_isolates(graph: nx.DiGraph) -> None:
    """Check for orphaned nodes.

    Args:
        graph: the graph to check

    """
    nodes = set(graph.nodes)
    root_node = next(nx.topological_sort(graph))

    isolates = list(set(nx.isolates(graph)) - {root_node})  # root can be isolate
    graph.remove_nodes_from(isolates)
    orphaned_nodes = nodes - set(graph.nodes)
    if orphaned_nodes:
        message = f"{orphaned_nodes} were isolates in the type relation map and consequently orphaned. "
        message += "Please add some mapping to the orphaned nodes."
        warnings.warn(message)


def check_cycles(graph: nx.DiGraph) -> None:
    """Check for cycles and warn if one is found

    Args:
        graph: the graph to check

    """
    cycles = list(nx.simple_cycles(graph))
    if len(cycles) > 0:
        warnings.warn(f"Cyclical relations between types {cycles} detected")


def traverse_graph_with_series(
    base_type: T,
    series: Sequence,
    graph: nx.DiGraph,
    path: List[T] = None,
    state: Optional[dict] = None,
) -> Tuple[Sequence, List[T], dict]:
    """Depth First Search traversal. There should be at most one successor that contains the series.

    Args:
        base_type: Entry-point for graph to start traversal
        series: the Series to check
        graph: the Graph to traverse
        path: the path so far
        state: traversal state

    Returns:
        The most uniquely specified node matching the series.
    """
    if state is None:
        state = dict()

    if path is None:
        path = []

    path.append(base_type)

    for vision_type in graph.successors(base_type):
        relation = graph[base_type][vision_type]["relationship"]

        if relation.is_relation(series, state):
            series = relation.transform(series, state)
            return traverse_graph_with_series(vision_type, series, graph, path, state)

    return series, path, state


def traverse_graph_with_sampled_series(
    base_type: T,
    series: pd.Series,
    graph: nx.DiGraph,
    sample_size: int = 10,
    state: dict = dict(),
) -> Tuple[Sequence, List[T], dict]:
    """Depth First Search traversal with sampling. There should be at most one successor that contains the series.

    Args:
        base_type: Entry-point for graph to start  traversal
        series: the Series to check
        graph: the Graph to traverse
        sample_size: number of items used in heuristic traversal
        state: traversal state

    Returns:
        The most uniquely specified node matching the series.
    """

    if (series.shape[0] < 1000) or (sample_size > series.shape[0]):
        return traverse_graph_with_series(base_type, series, graph, state=state)

    series_sample = series.sample(sample_size)
    _, path, _ = traverse_graph_with_series(
        base_type, series_sample, graph, state=state
    )
    if len(path) == 1:
        return series, path, state

    # Cast the full series
    from_type = path[0]
    for i, to_type in enumerate(path[1:]):
        relation = graph[from_type][to_type]["relationship"]
        if not relation.is_relation(series, state):
            break
        series = relation.transform(series, state)
        from_type = to_type

    return series, path[0 : (i + 2)], state


@singledispatch
def traverse_graph(
    data: Sequence, root_node: T, graph: nx.DiGraph
) -> Tuple[Sequence, Union[List[T], Dict[str, List[T]]], Dict[str, dict]]:
    return traverse_graph_with_series(root_node, data, graph)


@singledispatch
def get_type_from_path(
    path_data: Union[Sequence[T], Dict[str, Sequence[T]]]
) -> Union[T, Dict[str, T]]:
    raise TypeError(f"Can't get types from path object of type {type(path_data)}")


@get_type_from_path.register(list)
@get_type_from_path.register(tuple)
def _get_type_from_path_builtin(path_list: Sequence[T]) -> T:
    return path_list[-1]


@get_type_from_path.register(dict)
def _get_type_from_path_dict(path_dict: Dict[str, Sequence[T]]) -> Dict[str, T]:
    return {k: v[-1] for k, v in path_dict.items()}


class VisionsTypeset:
    """
    A collection of :class:`visions.types.type.VisionsBaseType` with  associated relationship map between them.

    Attributes:
        types: The collection of Visions Types derived from :class:`visions.types.type.VisionsBaseType`
        base_graph: The graph of relations composed exclusively of :class:`visions.relations.relations.IdentityRelation`
        relation_graph: The full relation graph including both :class:`visions.relations.relations.IdentityRelation`
            and :class:`visions.relations.relations.InferenceRelation`
    """

    def __init__(self, types: Set[Type[VisionsBaseType]]) -> None:
        """
        Args:
            types: a set of types
        """
        self._root_node: Optional[T] = None

        if not isinstance(types, Iterable):
            raise ValueError("types should be Sequence")

        self.relation_graph, self.base_graph = build_graph(set(types))

        if not issubclass(self.root_node, Generic):
            raise ValueError("`root_node` should be a subclass of Generic")

        self.types = set(self.relation_graph.nodes)

    @property
    def root_node(self) -> T:
        """Returns a cached copy of the relation_graphs root node

        Args:

        Returns:
            A cached copy of the relation_graphs root node.
        """
        if self._root_node is None:
            self._root_node = next(nx.topological_sort(self.relation_graph))
        return self._root_node

    def detect(self, data: Any) -> Tuple[Sequence, Any, dict]:
        """The results found after only considering IdentityRelations.

        Notes:
            This is an advanced feature, consider using `detect_type` in case the type is what is needed.

        Args:
            data: a DataFrame or Series to determine types over

        Returns:
            A tuple of the coerced sequence, visited nodes and state
        """
        return traverse_graph(data, self.root_node, self.base_graph)

    def detect_type(self, data: Sequence) -> Union[T, Dict[str, T]]:
        """The inferred type found only considering IdentityRelations.

        Args:
            data: a DataFrame or Series to determine types over

        Returns:
            A dictionary of {name: type} pairs in the case of DataFrame input or a type
        """
        _, paths, _ = self.detect(data)
        return get_type_from_path(paths)

    def infer(self, data: Sequence) -> Tuple[Sequence, Any, dict]:
        """The results found after considering all relations.

        Notes:
            This is an advanced feature, consider using `infer_type` in case the type is what is needed.

        Args:
            data: a DataFrame or Series to determine types over

        Returns:
            A tuple of the coerced sequence, visited nodes and state
        """
        return traverse_graph(data, self.root_node, self.relation_graph)

    def infer_type(self, data: Sequence) -> Union[T, Dict[str, T]]:
        """The inferred type found using all type relations.

        Args:
            data: a DataFrame or Series to determine types over

        Returns:
            A dictionary of {name: type} pairs in the case of DataFrame input or a type
        """
        _, paths, _ = self.infer(data)
        return get_type_from_path(paths)

    def cast_to_detected(self, data: Sequence) -> Sequence:
        """Transforms input data into a canonical representation using only IdentityRelations

        Args:
            data: a DataFrame or Series to determine types over

        Returns:
            new_data: The transformed DataFrame or Series.
        """
        data, _, _ = self.detect(data)
        return data

    def cast_to_inferred(self, data: Sequence) -> Sequence:
        """Transforms input data and returns it's corresponding new type relation using all relations.

        Args:
            data: a DataFrame or Series to determine types over

        Returns:
            new_data: The transformed DataFrame or Series.
            types: A dictionary of {name: type} pairs in the case of DataFrame input or a type.
        """
        data, _, _ = self.infer(data)
        return data

    def output_graph(
        self,
        file_name: Union[str, Path],
        base_only: bool = False,
        dpi: Optional[int] = None,
    ) -> None:
        """Write the type graph to a file.

        Args:
            file_name: the file to save the output to
            base_only: if True, plot the graph without relation mapping edges
            dpi: set the dpi of the output image
        """
        from visions.utils.graph import output_graph

        if base_only:
            graph = self.base_graph.copy()
        else:
            graph = self.relation_graph.copy()

        graph.graph["node"] = {"shape": "box", "color": "red"}
        if dpi is not None:
            graph.graph["graph"] = {"dpi": dpi}

        output_graph(graph, file_name)

    def plot_graph(
        self,
        dpi: int = 800,
        base_only: bool = False,
        figsize: Optional[Tuple[int, int]] = None,
    ):
        """

        Args:
            dpi: dpi of the matplotlib figure.
            figsize: figure size
            base_only: Only display the typesets base_graph
        Returns:
            Displays the image
        """
        import os
        import tempfile

        from matplotlib import image as mpimg
        from matplotlib import pyplot as plt

        with tempfile.NamedTemporaryFile(suffix=".png", delete=False) as temp_file:
            self.output_graph(temp_file.name, dpi=dpi, base_only=base_only)
            img = mpimg.imread(temp_file.name)
            plt.figure(dpi=dpi, figsize=figsize)
            plt.axis("off")
            plt.imshow(img)
        os.unlink(temp_file.name)

    def _get_other_type(self, other: TypeOrTypeset) -> Set[T]:
        """Converts input into a set of :class:`visions.types.type.VisionsBaseType`

        Args:
            other: A :class:`visions.types.type.VisionsBaseType` or :class:`visions.typesets.typeset.VisionsTypeset`

        Raises:
            NotImplementedError:

        Returns:
            Set[Type[VisionsBaseType]]:
        """
        if isinstance(other, VisionsTypeset):
            other_types = set(other.types)
        elif issubclass(other, VisionsBaseType):
            other_types = {other}
        else:
            raise NotImplementedError(
                f"Typeset operation not implemented for type {type(other)}"
            )
        return other_types

    def replace(self, old: T, new: T) -> "VisionsTypeset":
        """Create a new typeset having replace one type with another.

        Args:
            old: Visions type to replace.
            new: Replacement visions type.

        Returns
            A VisionsTypeset
        """
        types = self.types.copy()
        types.add(new)
        types.remove(old)
        return VisionsTypeset(types)

    def __add__(self, other: TypeOrTypeset) -> "VisionsTypeset":
        """Adds a type or typeset into the current typeset.

        Args:
            other: Type or typeset to be added

        Returns
            A VisionsTypeset
        """
        other_types = self._get_other_type(other)
        return VisionsTypeset(self.types | other_types)

    def __iadd__(self, other: TypeOrTypeset) -> "VisionsTypeset":
        """Adds a type or typeset into the current typeset.

        Args:
            other: Type or typeset to be added

        Returns
            A VisionsTypeset
        """
        return self.__add__(other)

    def __sub__(self, other: TypeOrTypeset) -> "VisionsTypeset":
        """Subtracts a type or typeset from the current typeset.

        Args:
            other: Type or typeset to be removed

        Returns
            A VisionsTypeset
        """
        other_types = self._get_other_type(other)
        return VisionsTypeset(self.types - other_types)

    def __isub__(self, other: TypeOrTypeset) -> "VisionsTypeset":
        """Subtracts a type or typeset from the current typeset.

        Args:
            other: Type or typeset to be removed

        Returns
            A VisionsTypeset
        """
        return self.__sub__(other)

    def __repr__(self) -> str:
        """Pretty representation of the typeset.

        Returns
            A :class:`visions.typesets.typeset.VisionsTypeset`
        """
        return self.__class__.__name__



from typing import Dict, List, Sequence, Tuple, Type, Union

import pandas as pd

from visions.types.type import VisionsBaseType
from visions.typesets.typeset import VisionsTypeset

T = Type[VisionsBaseType]


def cast_to_detected(data: Sequence, typeset: VisionsTypeset) -> Sequence:
    """Casts a DataFrame into a typeset by first performing column wise type detection against
    a provided typeset

    Args:
        data: the DataFrame to cast
        typeset: the Typeset in which we cast

    Returns:
        A tuple of the casted DataFrame and the types to which the columns were cast
    """
    return typeset.cast_to_detected(data)


def cast_to_inferred(data: Sequence, typeset: VisionsTypeset) -> Sequence:
    """Casts a DataFrame into a typeset by first performing column wise type inference against
    a provided typeset

    Args:
        data: the DataFrame to cast
        typeset: the Typeset in which we cast

    Returns:
        A tuple of the casted DataFrame and the types to which the columns were cast
    """
    return typeset.cast_to_inferred(data)


def infer_type(data: Sequence, typeset: VisionsTypeset) -> Union[Dict[str, T], T]:
    """Infer the current types of each column in the DataFrame given the typeset.

    Args:
        data: the DataFrame to infer types on
        typeset: the Typeset that provides the type context

    Returns:
        A dictionary with a mapping from column name to type
    """
    return typeset.infer_type(data)


def detect_type(data: Sequence, typeset: VisionsTypeset) -> Union[Dict[str, T], T]:
    """Detect the type in the base graph

    Args:
        data: the DataFrame to detect types on
        typeset: the Typeset that provides the type context

    Returns:
        A dictionary with a mapping from column name to type
    """
    return typeset.detect_type(data)


def compare_detect_inference_frame(
    data: Sequence, typeset: VisionsTypeset
) -> List[Tuple[str, T, T]]:
    """Compare the types given by inference on the base graph and the relational graph

    Args:
        data: the sequence to detect types on
        typeset: the Typeset that provides the type context

    Examples:
        >>> for column, type_before, type_after in compare_detect_inference_frame(data, typeset):
        >>>    print(f"{column} was {type_before} is {type_after}")

    See Also:
        :doc:`type_inference_report_frame <visions.functional.type_inference_report_frame>`:
            Formatted report of the output of this function
    """
    comparisons = []
    detected_types = detect_type(data, typeset)
    inferred_types = infer_type(data, typeset)

    assert isinstance(detected_types, dict) and isinstance(
        inferred_types, dict
    )  # Placate the MyPy Gods

    for key in detected_types.keys() & inferred_types.keys():
        comparisons.append((key, detected_types[key], inferred_types[key]))
    return comparisons


# TODO: make independent of pandas
def type_inference_report_frame(df: pd.DataFrame, typeset: VisionsTypeset) -> str:
    """Return formatted report of the output of `compare_detect_inference_frame`.

    Args:
        df: the DataFrame to detect types on
        typeset: the Typeset that provides the type context

    Returns:
        Text-based comparative type inference report

    Examples:
        >>> import pandas as pd
        >>> from visions.functional import type_inference_report_frame
        >>> from visions.typesets import StandardSet
        >>>
        >>> typeset = StandardSet()
        >>> df = pd.read_csv('dataset.csv')
        >>>
        >>> report = type_inference_report_frame(df, typeset)
        >>> print(report)
    """
    padding = 5
    max_column_length = max(len(column) for column in df.columns) + padding
    max_type_length = 30

    report = ""
    change_count = 0
    for column, type_before, type_after in compare_detect_inference_frame(df, typeset):
        changed = type_before != type_after
        if changed:
            fill = "!="
            change_count += 1
        else:
            fill = "=="
        report += (
            f"{column: <{max_column_length}} {type_before: <{max_type_length}} "
            f"{fill} "
            f"{type_after: <{max_type_length}} \n"
        )
    report += (
        "In total {change_count} out of {type_count} types were changed.\n".format(
            change_count=change_count, type_count=len(df.columns)
        )
    )
    return report



import pandas as pd
from pandas.api import types as pdt

from visions.backends.pandas.series_utils import series_not_empty, series_not_sparse
from visions.types.time_delta import TimeDelta


@TimeDelta.contains_op.register
@series_not_sparse
@series_not_empty
def time_delta_contains(series: pd.Series, state: dict) -> bool:
    """
    Example:
        >>> x = pd.Series([pd.Timedelta(days=i) for i in range(3)])
        >>> x in visions.Timedelta
        True
    """
    return pdt.is_timedelta64_dtype(series)



import pandas as pd
from pandas.api import types as pdt

from visions.backends.pandas import test_utils
from visions.backends.pandas.series_utils import (
    series_handle_nulls,
    series_not_empty,
    series_not_sparse,
)
from visions.types import DateTime, String


def pandas_infer_datetime(series: pd.Series, state: dict) -> pd.Series:
    try:
        return pd.to_datetime(series)
    except Exception:
        pass

    return pd.to_datetime(series, format="mixed")


@DateTime.register_relationship(String, pd.Series)
@series_handle_nulls
def string_is_datetime(series: pd.Series, state: dict) -> bool:
    def string_to_datetime_func(series: pd.Series) -> pd.Series:
        return pandas_infer_datetime(series, state)

    exceptions = [OverflowError, TypeError]
    coerced_series = test_utils.option_coercion_evaluator(
        string_to_datetime_func, exceptions
    )(series)

    if coerced_series is None:
        return False
    else:
        return not coerced_series.dropna().empty


@DateTime.register_transformer(String, pd.Series)
def string_to_datetime(series: pd.Series, state: dict) -> pd.Series:
    return pandas_infer_datetime(series, state)


@DateTime.contains_op.register
@series_not_sparse
@series_handle_nulls
@series_not_empty
def datetime_contains(series: pd.Series, state: dict) -> bool:
    return pdt.is_datetime64_any_dtype(series)



import math
from typing import Union

import numpy as np
import pandas as pd
from pandas.api import types as pdt

from visions.backends.pandas import test_utils
from visions.backends.pandas.series_utils import series_not_empty, series_not_sparse
from visions.backends.shared.parallelization_engines import pandas_apply
from visions.types.complex import Complex
from visions.types.string import String


def imaginary_in_string(
    series: pd.Series, imaginary_indicator: tuple = ("j", "i")
) -> bool:
    return any(any(v in s for v in imaginary_indicator) for s in series)


def convert_val_to_complex(val: str) -> Union[complex, float]:
    result = complex(val)
    return (
        np.nan if any(math.isnan(val) for val in (result.real, result.imag)) else result
    )


def convert_to_complex_series(series: pd.Series) -> pd.Series:
    return pandas_apply(series, convert_val_to_complex)


@Complex.register_relationship(String, pd.Series)
def string_is_complex(series: pd.Series, state: dict) -> bool:
    coerced_series = test_utils.option_coercion_evaluator(convert_to_complex_series)(
        series
    )

    return (
        coerced_series is not None
        and not all(v.imag == 0 for v in coerced_series.dropna())
        and imaginary_in_string(series)
    )


@Complex.register_transformer(String, pd.Series)
def string_to_complex(series: pd.Series, state: dict) -> pd.Series:
    return convert_to_complex_series(series)


@Complex.contains_op.register
@series_not_sparse
@series_not_empty
def complex_contains(series: pd.Series, state: dict) -> bool:
    return pdt.is_complex_dtype(series)



import functools
from collections import OrderedDict

import pandas as pd


class LRUCacher:
    def __init__(self, hash_func, max_length, value_func):
        self.hash_func = hash_func
        self.max_length = max_length
        self.value_func = value_func
        self.cache = OrderedDict()

    def __getitem__(self, key):
        value = self.cache[key]
        self.cache.move_to_end(key)
        return value

    def __setitem__(self, key, value):
        if key in self.cache:
            self.cache.move_to_end(key)
        self.cache[key] = value
        if len(self.cache) > self.max_length:
            oldest = next(iter(self.cache))
            del self.cache[oldest]

    def get_key(self, *args):
        return self.hash_func(*args)

    def get(self, *args):
        id_key = self.get_key(*args)
        if id_key not in self.cache:
            self[id_key] = self.value_func(*args)
        return self[id_key]


def lru_cache(hash_func, max_length):
    def func_inner(func):
        cache = LRUCacher(hash_func, max_length, func)

        @functools.wraps(func)
        def inner(*args):
            return cache.get(*args)

        return inner

    return func_inner


def mutable_pseudo_hash(data, node, graph):
    # return id((data, node, graph))
    try:
        if isinstance(data, pd.DataFrame):
            data_hash = hash(hash(tuple(data[col])) for col in data.columns)
        else:
            data_hash = hash(tuple(data.values))
    except (ValueError, TypeError, AttributeError):
        return id((data, node, graph))

    return hash((data_hash, node, graph))



from typing import Dict

import pandas as pd


def get_series() -> Dict[str, pd.Series]:
    from visions.backends.numpy.sequences import get_sequences as get_numpy_sequences
    from visions.backends.pandas.sequences import get_sequences as get_pandas_sequences
    from visions.backends.python.sequences import get_sequences as get_builtin_sequences

    sequences = get_builtin_sequences()
    sequences.update(get_numpy_sequences())

    test_series = {}
    for name, sequence in sequences.items():
        test_series[name] = pd.Series(sequence)

    test_series.update(get_pandas_sequences())
    assert all(isinstance(v, pd.Series) for v in test_series.values())

    return test_series



import pathlib
from typing import Sequence

from visions.types.file import File


@File.contains_op.register
def file_contains(sequence: Sequence, state: dict) -> bool:
    return all(isinstance(p, pathlib.Path) and p.exists() for p in sequence)



import pandas as pd
from pandas.api import types as pdt

from visions.types.sparse import Sparse


@Sparse.contains_op.register
def sparse_contains(series: pd.Series, state: dict) -> bool:
    return pdt.is_sparse(series)



import math
from datetime import datetime, timedelta

import numpy as np
import pandas as pd

from .utilities import has_import

has_numba = has_import("numba")

if has_numba:
    import numba as nb


def nan_mask(array: np.ndarray) -> np.ndarray:
    # TODO: Fails for values like None, pandas resolves this but it's complicated some links:
    # https://github.com/pandas-dev/pandas/blob/3391a348f3f7cd07a96c8e6a4b05e3e9f60c8567/pandas/core/series.py#L192
    # https://github.com/pandas-dev/pandas/blob/65319af6e563ccbb02fb5152949957b6aef570ef/pandas/core/base.py#L816
    # https://github.com/pandas-dev/pandas/blob/65319af6e563ccbb02fb5152949957b6aef570ef/pandas/core/dtypes/missing.py#L133
    # https://github.com/pandas-dev/pandas/blob/65319af6e563ccbb02fb5152949957b6aef570ef/pandas/core/dtypes/missing.py#L202
    # raise NotImplementedError('Robust missing value detection not implemented for numpy arrays')
    try:
        mask = ~np.isnan(array)
    except TypeError:
        # mask = np.array([not pd. for v in array], dtype=bool)
        mask = ~pd.isna(array)
    return mask


# TODO: There are optimizations here, just have to define precisely the desired missing ruleset in the
# generated jit
if has_numba:

    def is_missing(x):
        """
        Return True if the value is missing, False otherwise.
        """
        if isinstance(x, float):
            return np.isnan(x)
        elif isinstance(x, (datetime, timedelta)):
            missing = x("NaT")
            return x == missing
        elif x is None:
            return True
        else:
            return False

    nb.extending.overload(is_missing)(lambda x: is_missing)

    @nb.jit(nopython=True)
    def hasna(x: np.ndarray) -> bool:
        for item in x:
            if is_missing(item):
                return True
        return False

else:

    def anynan(array: np.ndarray) -> bool:
        return any(math.isnan(v) for v in array)



from typing import Sequence
from urllib.parse import ParseResult, urlparse

from visions.types.string import String
from visions.types.url import URL


@URL.contains_op.register
def url_contains(sequence: Sequence, state: dict) -> bool:
    return all(isinstance(value, ParseResult) for value in sequence)


@URL.register_transformer(String, Sequence)
def string_to_url(sequence: Sequence, state: dict) -> Sequence:
    return tuple(map(urlparse, sequence))


@URL.register_relationship(String, Sequence)
def string_is_url(sequence: Sequence, state: dict) -> bool:
    try:
        return all(x.netloc and x.scheme for x in string_to_url(sequence, {}))
    except (ValueError, TypeError, AttributeError):
        return False
